#!/usr/bin/env python

import os
import sys
import logging
import signal
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from functools import partial
from getpass import getpass, getuser
from json import dumps
from os.path import abspath
from socket import create_connection
from textwrap import dedent
from threading import Thread
from time import sleep
from typing import Any, Dict, List, Tuple

from paramiko import Channel, Transport, SSHConfig
from paramiko import RSAKey, DSSKey, ECDSAKey, Ed25519Key
from paramiko import SSHException, ChannelException, PasswordRequiredException

from .utils import SERVER_ADDR, SSH_MUX, TIMEOUT
from .utils import Flag, Packet, send_pkt, recv_pkt
from .utils import Sender, Receiver, progress


class Client:
    default_port = 22
    default_dir = os.path.expanduser('~/.ssh')
    default_config = os.path.join(default_dir, 'config')

    def __init__(self, cli_parser: ArgumentParser):
        args = cli_parser.parse_args()

        # init logger
        self.log_level = {
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
            3: logging.DEBUG
        }.get(args.verbose, logging.ERROR)
        self.set_log()

        try:
            # init instance attributes from CLI args:
            #   - self.action:   Flag.PUSH | Flag.PULL
            #   - self.host:     ssh server's ip or domain
            #   - self.username: username used to login to the ssh server
            #   - self.srcs:     source dirs or files
            #   - self.dst:      dest dir or file
            self.parse_cli_args(args)
        except Exception as e:
            logging.error(f'fcp: {e}')
            print('--------------------------------')
            cli_parser.print_help()
            sys.exit(1)

        self.config = self.load_ssh_config(self.host, args.ssh_config)
        self.n_tunnel = args.num
        self.n_channel = self.n_tunnel * SSH_MUX
        self.port = args.port
        self.pkey_path = args.private_key
        self.include = args.include
        self.exclude = [p for p in args.exclude.split(',') if p]

        # the ssh tunnels
        self.tunnels: Dict[Transport, List[Channel]] = {}

    @staticmethod
    def parse_remote_addr(remote):
        '''解析远程主机登录信息'''
        netloc, path = remote.split(':')
        user, host = netloc.split('@') if '@' in netloc else ('', netloc)
        return user, host, path

    @classmethod
    def parse_remote_sources(cls, sources):
        users, hosts, srcs = set(), set(), set()
        for src in sources:
            user, host, path = cls.parse_remote_addr(src)
            users.add(user)
            hosts.add(host)
            srcs.add(path)
        if len(users) == 1 and len(hosts) == 1:
            return users.pop(), hosts.pop(), sorted(srcs)
        else:
            raise ValueError('All source args must come from '
                             'the same machine with same user.')

    def parse_cli_args(self, args):
        '''解析命令行参数'''
        if ':' in args.srcs[0]:
            self.username, self.host, self.srcs = self.parse_remote_sources(args.srcs)
            self.action = Flag.PULL
            self.dst = abspath(args.dst)
        elif ':' in args.dst:
            self.username, self.host, self.dst = self.parse_remote_addr(args.dst)
            self.action = Flag.PUSH
            self.srcs = [abspath(path) for path in args.srcs]
        else:
            raise ValueError('The server address is not specified.')

    def set_log(self):
        '''处理日志'''
        global print
        print = partial(progress.print, style='blue')

        if self.log_level <= logging.ERROR:
            logging.error = partial(progress.print, style='red')
        if self.log_level <= logging.WARNING:
            logging.warning = partial(progress.print, style='yellow')
        if self.log_level <= logging.INFO:
            logging.info = partial(progress.print, style='blue')
        if self.log_level <= logging.DEBUG:
            logging.debug = partial(progress.print, style='white')

        print = partial(progress.print, style='blue')
        paramiko_logger = logging.getLogger("paramiko")
        paramiko_logger.setLevel(logging.ERROR)

    @classmethod
    def load_ssh_config(cls, hostname: str, user_config_file=None) -> dict:
        '''加载默认配置'''
        _path = user_config_file or cls.default_config
        cfg = SSHConfig.from_path(_path)
        return cfg.lookup(hostname)

    @staticmethod
    def load_pkey(key_path):
        '''加载 Key'''
        _path = os.path.abspath(os.path.expanduser(key_path))
        if not os.path.isfile(_path):
            return

        # guess key type
        filename = os.path.basename(_path)
        key_type = filename.split('_')[1] if filename.startswith('id_') else ''
        key_types = ['rsa', 'ed25519', 'dsa', 'ecdsa']
        key_classes = {
            'rsa': RSAKey,
            'ed25519': Ed25519Key,
            'dsa': DSSKey,
            'ecdsa': ECDSAKey
        }

        types_to_try = [key_type] if key_type in key_types else key_types
        for _ in range(3):
            password = None
            for _type in types_to_try:
                key_cls = key_classes[_type]
                try:
                    return key_cls.from_private_key_file(key_path, password)
                except PasswordRequiredException:
                    password = getpass(f"Enter password for key '{key_path}': ")
                except SSHException:
                    continue

    def search_pkeys(self):
        '''查找可用的私钥'''
        if self.pkey_path:
            pkey_paths = [self.pkey_path]
        else:
            pkey_paths = self.config.get('identityfile', [])

        if not pkey_paths:
            for keyname in os.listdir(self.default_dir):
                if keyname.startswith('id_') and not keyname.endswith('.pub'):
                    path = os.path.join(self.default_dir, keyname)
                    pkey_paths.append(path)

        logging.debug(f'found pkeys: {pkey_paths}')

        return pkey_paths

    def create_transport(self, sock, user, pkey, password):
        if isinstance(sock, tuple):
            sock = create_connection(sock)
        tp = Transport(sock)
        tp.use_compression(True)
        tp.set_keepalive(60)
        try:
            tp.connect(username=user, pkey=pkey, password=password)
            if tp.is_authenticated():
                self.tunnels[tp] = []
                return tp
        except SSHException as e:
            tp.stop_thread()
            logging.error(f'fcp: {e}')

    def create_channel(self, transport: Transport) -> Channel:
        '''create a new channel by transport'''
        try:
            # create channel
            channel = transport.open_channel(kind='direct-tcpip',
                                             dest_addr=SERVER_ADDR,
                                             src_addr=('localhost', 0))
            channel.settimeout(TIMEOUT)

            # add to self.tunnels
            channels = self.tunnels.setdefault(transport, [])
            channels.append(channel)

            return channel
        except (SSHException, ChannelException) as e:
            logging.error(f'fcp: {e}')
            sys.exit(1)

    def ssh_connect(self) -> Tuple[Transport, Any, Any]:  # type: ignore
        '''连接 SSH 服务器'''
        # 获取 SSH 服务器的连接参数
        self.host = self.config.get('hostname', self.host)
        self.port = self.port or self.config.get('port') or self.default_port
        self.port = int(self.port)
        self.username = self.username or self.config.get('user') or getuser()
        logging.debug(f'login info: {self.username}@{self.host}:{self.port}')

        # search private keys
        pkey_paths = self.search_pkeys()

        # connect to ssh server (just connect, not auth)
        addr = (self.host, self.port)
        sock = create_connection(addr)

        # try the pkeys one by one
        for _path in pkey_paths:
            logging.debug(f'test pkey: {_path}')
            pkey = self.load_pkey(_path)
            tp = self.create_transport(sock, self.username, pkey, None)
            if tp:
                self.tunnels[tp] = []
                return tp, pkey, None

        # try to auth with password
        for _ in range(3):
            password = getpass(f'password for {self.username}@{self.host}: ')
            tp = self.create_transport(sock, self.username, None, password)
            if tp:
                self.tunnels[tp] = []
                return tp, None, password

        logging.error('fcp: failed to create SSH tunnel')
        sys.exit(1)

    def handshake(self, channel, remote_path: str):
        '''握手'''
        conn_pkt = Packet.load(self.action, remote_path)
        send_pkt(channel, conn_pkt)
        session_pkt = recv_pkt(channel)
        session_id, = session_pkt.unpack_body()
        logging.info(f'[cyan]fcp[/cyan]: Channel-{id(channel):x} connected')

        return session_id

    def create_attached_channels(self, tp, conn_pool, session_id):
        channels = self.tunnels[tp]
        attach_pkt = Packet.load(Flag.ATTACH, session_id)

        def _attache_channel():
            channel = self.create_channel(tp)
            send_pkt(channel, attach_pkt)
            conn_pool.add(channel)
            logging.info(f'[cyan]fcp[/cyan]: Channel-{id(channel):x} connected')

        for _ in range(SSH_MUX - len(channels)):
            thr = Thread(target=_attache_channel, daemon=True)
            thr.start()

    def attached_connect(self, conn_pool, session_id, pkey, password):
        '''后续连接'''
        addr = (self.host, self.port)

        def _attache():
            tp = self.create_transport(addr, self.username, pkey, password)
            self.create_attached_channels(tp, conn_pool, session_id)

        # create channels from exists transports
        for tp in self.tunnels:
            self.create_attached_channels(tp, conn_pool, session_id)

        # create new transports
        for _ in range(self.n_tunnel - len(self.tunnels)):
            thr = Thread(target=_attache,
                         daemon=True)
            thr.start()
            sleep(0.1)  # 并发连接同一SSH服务器可能会报错，所以加一点延迟

    def start(self):
        try:
            progress.start()
            print('[cyan]fcp[/cyan]: connecting to server')
            tp, pkey, password = self.ssh_connect()
            first_channel = self.create_channel(tp)

            if self.action == Flag.PULL:
                remote_path = dumps({
                    'srcs': self.srcs,
                    'include': self.include,
                    'exclude': self.exclude
                }, ensure_ascii=False, separators=(',', ':'))
                session_id = self.handshake(first_channel, remote_path)
                porter = Receiver(session_id, self.dst, self.n_channel)
                print('[cyan]fcp[/cyan]: receiving files')
            else:
                session_id = self.handshake(first_channel, self.dst)
                porter = Sender(session_id, self.srcs, self.n_channel,
                                self.include, self.exclude)
                print('[cyan]fcp[/cyan]: sending files')

            porter.conn_pool.add(first_channel)
            porter.start()

            # create attached connections
            t = Thread(target=self.attached_connect,
                       args=(porter.conn_pool, session_id, pkey, password))
            t.start()

            porter.join()
            print('[cyan]fcp[/cyan]: [bold green]finished![/bold green]')
        except Exception as e:
            logging.error(f'fcp: {e}')
            if self.log_level == logging.DEBUG:
                progress.console.print_exception()
            sys.exit(1)
        finally:
            progress.stop()


def handle_sigint(signum, frame):
    '''键盘中断事件的处理'''
    logging.error('fcp: user canceled.')
    progress.stop()
    sys.exit(1)


signal.signal(signal.SIGINT, handle_sigint)


def main():
    parser = ArgumentParser(
        prog='fcp',
        formatter_class=RawDescriptionHelpFormatter,
        description=dedent('''
            PULL : fcp [OPTIONS...] [USER@]HOST:SRC... DST
            PUSH : fcp [OPTIONS...] SRC... [USER@]HOST:DST
        ''')
    )

    parser.add_argument('-p', dest='port', type=int, default=None,
                        help='The port of SSH server (default: 22)')

    parser.add_argument('-i', dest='private_key', type=str, default=None,
                        help='The private key file for SSH')

    parser.add_argument('-F', dest='ssh_config', type=str, default=None,
                        help='The config file for SSH (default: ~/.ssh/config)')

    parser.add_argument('-n', dest='num', type=int, default=8,
                        help='Max number of SSH tunnels (default: %(default)s)')

    parser.add_argument('-v', dest='verbose', action='count', default=0,
                        help='Verbose mode (default: disable)')

    parser.add_argument('--include', type=str, metavar='PATTERN', default='*',
                        help='include files matching PATTERN')

    parser.add_argument('--exclude', type=str, metavar='PATTERN', default='',
                        help='exclude files matching PATTERN, split by `,`')

    parser.add_argument(dest='srcs', nargs='+', help='source path')
    parser.add_argument(dest='dst', help='destination path')

    cli = Client(parser)
    cli.start()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import _socket
import logging
from argparse import ArgumentParser
from json import loads
from socket import socket
from socket import error as SocketError, timeout as TimeoutError
from socket import AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR, SO_REUSEPORT
from threading import Lock, Thread
from typing import Dict
from uuid import uuid4

import daemon

from .utils import Flag, SERVER_ADDR
from .utils import Packet, send_pkt, recv_pkt
from .utils import Sender, Receiver, Porter


class WatchDog(Thread):
    def __init__(self, server: 'Server', sock: socket):
        super().__init__(daemon=True)
        self.server = server
        self.sock = sock

    def run(self):
        try:
            # 等待接收新连接的第一个数据报文
            logging.debug('[WatchDog] waiting for handshake from %s:%d'
                          % self.sock.getpeername())
            self.sock.settimeout(60)
            packet = recv_pkt(self.sock)
            self.sock.settimeout(None)
        except ConnectionResetError:
            logging.error('[WatchDog] connection reset by peer.')
            return
        except TimeoutError:
            # 超时退出
            logging.error('[WatchDog] handshake timeout.')
            self.sock.close()
            return

        if packet.flag == Flag.PULL or packet.flag == Flag.PUSH:
            # 创建 Porter
            path, = packet.unpack_body()
            porter = self.server.create_porter(packet.flag, path)
            porter.conn_pool.add(self.sock)
            porter.start()

            # 将 SID 发送给客户端
            packet = Packet.load(Flag.SID, porter.sid)
            send_pkt(self.sock, packet)

        elif packet.flag == Flag.ATTACH:
            sid, = packet.unpack_body()
            if not self.server.porters[sid].conn_pool.add(self.sock):
                self.sock.close()

        else:
            # 对于错误的类型，直接关闭连接
            logging.debug('close conn')
            self.sock.close()


class Server(Thread):
    max_tasks = 256  # 同时运行的最大任务数量

    def __init__(self, max_conn) -> None:
        super().__init__(daemon=True)
        self.addr = SERVER_ADDR
        self.max_conn = max_conn  # 一个 Porter 的最大连接数
        self.is_running = True
        self.mutex = Lock()
        self.porters: Dict[bytes, Porter] = {}

    def create_porter(self, cli_flag: Flag, path: str) -> Porter:
        '''创建新 Porter'''
        sid = uuid4().bytes
        if cli_flag == Flag.PULL:
            _path = loads(path)
            srcs = _path['srcs']
            include = _path['include']
            exclude = _path['exclude']
            logging.debug(f'[Server] New task-{sid.hex()} for send {path}')
            self.porters[sid] = Sender(sid, srcs, self.max_conn,
                                       include, exclude)
        else:
            logging.debug(f'[Server] New task-{sid.hex()} for recv {path}')
            self.porters[sid] = Receiver(sid, path, self.max_conn)
        return self.porters[sid]

    def close_all_porters(self):
        '''关闭所有 Porter'''
        logging.debug('[Server] Closing all porters.')
        for porter in self.porters.values():
            porter.close()

    @staticmethod
    def create_socket_server(address, *, family=AF_INET, backlog=None,
                             reuse_port=False):
        """copyed from socket.py"""
        if reuse_port and not hasattr(_socket, "SO_REUSEPORT"):
            raise ValueError("SO_REUSEPORT not supported on this platform")

        sock = socket(family, SOCK_STREAM)
        try:
            if hasattr(_socket, 'SO_REUSEADDR'):
                try:
                    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
                except SocketError:
                    # Fail later on bind(), for platforms which may not
                    # support this option.
                    pass
            if reuse_port:
                sock.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
            try:
                sock.bind(address)
            except SocketError as err:
                msg = '%s (while attempting to bind on address %r)' % \
                    (err.strerror, address)
                raise SocketError(err.errno, msg) from None
            if backlog is None:
                sock.listen()
            else:
                sock.listen(backlog)
            return sock
        except SocketError:
            sock.close()
            raise

    def run(self):
        self.srv_sock = self.create_socket_server(self.addr,
                                                  backlog=2048,
                                                  reuse_port=True)
        logging.info('[Server] Listening to %s:%d' % self.addr)
        while self.is_running:
            # wait for new connection
            cli_sock, cli_addr = self.srv_sock.accept()
            logging.info('[Server] Accept new connection: %s:%s' % cli_addr)

            # create a WatchDog for handshake
            dog = WatchDog(self, cli_sock)
            dog.start()


def main():
    parser = ArgumentParser()
    parser.add_argument('-d',
                        dest='daemon',
                        action='store_true',
                        help='daemonize the fcp process.')

    parser.add_argument('-c',
                        dest='concurrency',
                        metavar='NUM',
                        type=int,
                        default=128,
                        help='max concurrent connections of one task.')

    parser.add_argument('--loglevel',
                        metavar='LEVEL',
                        default='error',
                        choices=['debug', 'info', 'warning', 'error'],
                        help=('specify the server verbosity level. '
                              'Choices: debug | info | warning | error'))

    args = parser.parse_args()

    loglevel = getattr(logging, args.loglevel.upper())
    logformat = '%(asctime)s %(levelname)7s %(module)s.%(lineno)s: %(message)s'

    if args.daemon:
        with daemon.DaemonContext():
            logging.basicConfig(filename='/tmp/fcpd.log',
                                level=loglevel,
                                datefmt='%Y-%m-%d %H:%M:%S',
                                format=logformat)
            server = Server(args.concurrency)
            server.start()
            server.join()
    else:
        logging.basicConfig(level=loglevel,
                            datefmt='%Y-%m-%d %H:%M:%S',
                            format=logformat)
        server = Server(args.concurrency)
        server.start()
        server.join()


if __name__ == '__main__':
    main()



import os
import sys
import shutil
import string
import random
from pathlib import Path
from queue import Queue
from unittest import TestCase, main

_project_dir = Path(__file__).parent.parent.as_posix()
sys.path.insert(0, _project_dir)
from filemanage import Reader


class TestReader(TestCase):
    def setUp(self) -> None:
        self.home = Path.home()  # 家目录
        self.local = Path(__file__).parent  # 测试目录
        self.testdir = self.local.joinpath('.fake')

        os.environ['TESTDIR'] = str(self.testdir)
        self.reader = Reader('.', Queue(), Queue())
        self.create_dirs_and_files()

    @staticmethod
    def rand_text(length):
        words = []
        while length > 1:
            n_chars = random.randint(3, 6) if length > 6 else length
            word = ''.join(random.choices(string.ascii_letters + '\n', k=n_chars))
            words.append(word)
            length = length - n_chars + 1
        return ' '.join(words)

    def create_dirs_and_files(self):
        print('create dirs')
        self.testdir.joinpath('a/b/c').mkdir(0o755, True, True)
        self.testdir.joinpath('x/y/z').mkdir(0o755, True, True)

        print('create files')
        files = ['f1', 'a/f2', 'a/f3', 'a/b/f4',
                 'a/b/f5', 'a/b/c/f6', 'a/b/c/f7',
                 'x/f8', 'x/y/f9', 'x/y/z/f10']
        for f in files:
            fpath = self.testdir.joinpath(f)
            fpath.touch(exist_ok=True)
            length = random.randint(1024, 1024 * 50)
            text = self.rand_text(length)
            fpath.write_text(text)

    def tearDown(self) -> None:
        shutil.rmtree(self.testdir)

    def test_abspath(self):
        print('\ntesting abspath(...)')
        self.assertEqual(Reader.abspath('/foo/bar/xy').as_posix(), '/foo/bar/xy')
        self.assertEqual(Reader.abspath('/[fp]o?/ba*').as_posix(), '/[fp]o?/ba*')

        self.assertEqual(Reader.abspath('~/foo/bar/xy').as_posix(), str(self.home.joinpath('foo/bar/xy')))
        self.assertEqual(Reader.abspath('~/[fp]o?/ba*').as_posix(), str(self.home.joinpath('[fp]o?/ba*')))

        self.assertEqual(Reader.abspath('$TESTDIR/foo/bar/xy').as_posix(), str(self.testdir.joinpath('foo/bar/xy')))
        self.assertEqual(Reader.abspath('$TESTDIR/[fp]o?/ba*').as_posix(), str(self.testdir.joinpath('[fp]o?/ba*')))

        self.assertEqual(Reader.abspath('foo/bar/xy').as_posix(), str(self.home.joinpath('foo/bar/xy')))
        self.assertEqual(Reader.abspath('[fp]o?/ba*').as_posix(), str(self.home.joinpath('[fp]o?/ba*')))

    def test_search_files_and_dirs(self):
        print('\ntesting search_files_and_dirs(...)')

        for target_path in ['$TESTDIR/a', '$TESTDIR/*', '~/src/fcp/*/*.py']:
            print(f'\n{target_path=}')
            for path, relpath in Reader.search_files_and_dirs(target_path):
                p1, p2 = str(path), str(relpath)
                self.assertTrue(p1.startswith('/'))
                self.assertTrue(p1.endswith(p2))
                print(f'{path.as_posix():40s}  ->  {relpath}')

        print("\ntarget_path='~/src/fcp/', filter='f*'")
        for path, relpath in Reader.search_files_and_dirs('~/src/fcp', 'f*'):
            p1, p2 = str(path), str(relpath)
            self.assertTrue(p1.startswith('/'))
            self.assertTrue(p1.endswith(p2))
            print(f'{path.as_posix():40s}  ->  {relpath}')


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as f_readme:
    long_description = f_readme.read()

setuptools.setup(
    name="fastcopy",
    version="0.1.4",
    python_requires=">=3.6",
    author="Seamile",
    author_email="lanhuermao@gmail.com",
    description="A multi-threaded file transfer tool over SSH. The goal is to replace SCP.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seamile/FastCopy",
    packages=['fastcopy'],
    install_requires=[
        "paramiko>=2.7.2",
        "python-daemon>=2.3.0",
        "rich>=10.6.0"
    ],
    entry_points={
        'console_scripts': [
            'fcp=fastcopy.client:main',
            'fcpd=fastcopy.server:main'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Topic :: Internet',
        'Topic :: Utilities',
    ],
)



import os
import re
import logging
from binascii import crc32
from collections import deque
from enum import IntEnum
from glob import has_magic, iglob
from hashlib import md5
from math import ceil
from paramiko import Channel
from pathlib import Path
from queue import Queue, Empty
from selectors import SelectSelector, EVENT_WRITE
from socket import socket, error as SocketError
from struct import pack, unpack
from threading import Event, Semaphore, Thread
from typing import (Any, Deque, Dict, Generator, Iterable, List, NamedTuple,
                    Set, Tuple, Union)

from rich.progress import (BarColumn, Progress, TaskID, TextColumn,
                           TransferSpeedColumn)


SERVER_ADDR = ('127.0.0.1', 7523)
CHUNK_SIZE = 1024 * 16  # 默认数据块大小 (单位: 字节)
SSH_MUX = 3
TIMEOUT = 60 * 5  # 全局超时时间
LEN_HEAD = 7

Connection = Union[socket, Channel]

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}"),
    BarColumn(bar_width=60),
    TransferSpeedColumn(),
    "•",
    "[progress.percentage]{task.percentage:>3.1f}%"
)


class Flag(IntEnum):
    PUSH = 1         # 推送申请
    PULL = 2         # 拉取申请
    SID = 3          # 建立会话
    ATTACH = 4       # 后续连接
    MONOFILE = 5     # 传输模式
    DIR_INFO = 6     # 目录信息
    FILE_INFO = 7    # 文件信息
    FILE_COUNT = 8   # 文件数量
    FILE_READY = 9   # 文件就绪
    FILE_CHUNK = 10  # 数据传输
    DONE = 11        # 完成
    EXCEPTION = 12   # 异常退出

    @classmethod
    def contains(cls, member: object) -> bool:
        return member in cls.__members__.values()


class Packet(NamedTuple):
    flag: Flag
    body: bytes

    def __str__(self) -> str:
        return (f'Packet: {self.flag.name} '
                f'len={self.length} '
                f'chk={self.chksum:08x}')

    @property
    def length(self) -> int:
        return len(self.body)

    @property
    def chksum(self) -> int:
        return crc32(self.body)

    @staticmethod
    def load(flag: Flag, *args) -> 'Packet':
        '''将包体封包'''
        if flag == Flag.PULL or flag == Flag.PUSH:
            if isinstance(args[0], bytes):
                body = args[0]
            else:
                body = str(args[0]).encode('utf8')
        elif flag == Flag.SID or flag == Flag.ATTACH:
            body = pack('>16s', *args)
        elif flag == Flag.MONOFILE:
            body = pack('>?', *args)
        elif flag == Flag.DIR_INFO:
            length = len(args[-1])
            body = pack(f'>IH{length}s', *args)
        elif flag == Flag.FILE_INFO:
            length = len(args[-1])
            body = pack(f'>IHQd16s{length}s', *args)
        elif flag == Flag.FILE_COUNT:
            body = pack('>I', *args)
        elif flag == Flag.FILE_READY:
            body = pack('>I', *args)
        elif flag == Flag.FILE_CHUNK:
            length = len(args[-1])
            body = pack(f'>2I{length}s', *args)
        elif flag == Flag.DONE:
            body = pack('>?', True)
        elif flag == Flag.EXCEPTION:
            body = str(args[0]).encode('utf8')
        else:
            raise ValueError(f'{flag} is not a valid Flag')
        return Packet(flag, body)

    def pack(self) -> bytes:
        '''封包'''
        fmt = f'>BIH{self.length}s'
        return pack(fmt, self.flag, self.chksum, self.length, self.body)

    @staticmethod
    def unpack_head(head: bytes) -> Tuple[Flag, int, int]:
        '''解析 head'''
        flag, chksum, length = unpack('>BIH', head)
        if not Flag.contains(flag):
            raise PacketError
        else:
            return Flag(flag), chksum, length

    def unpack_body(self) -> Tuple[Any, ...]:
        '''将 body 解包'''
        if self.flag == Flag.PULL or self.flag == Flag.PUSH:
            return (self.body.decode('utf-8'),)  # dest path

        elif self.flag == Flag.SID or self.flag == Flag.ATTACH:
            return unpack('>16s', self.body)  # Worker ID

        elif self.flag == Flag.MONOFILE:
            return unpack('>?', self.body)  # is monofile

        elif self.flag == Flag.DIR_INFO:
            # file_id | perm | path
            #   4B    |  2B  |  ...
            fmt = f'>IH{self.length - 6}s'
            return unpack(fmt, self.body)

        elif self.flag == Flag.FILE_INFO:
            # file_id | perm | size | mtime | chksum | path
            #   4B    |  2B  |  8B  |  8B   |  16B   |  ...
            fmt = f'>IHQd16s{self.length - 38}s'
            return unpack(fmt, self.body)

        elif self.flag == Flag.FILE_COUNT:
            return unpack('>I', self.body)  # file count

        elif self.flag == Flag.FILE_READY:
            return unpack('>I', self.body)  # file id

        elif self.flag == Flag.FILE_CHUNK:
            # file_id |  seq  | chunk
            #    4B   |  4B   |  ...
            fmt = f'>2I{self.length - 8}s'
            return unpack(fmt, self.body)

        elif self.flag == Flag.DONE:
            return unpack('>?', self.body)

        elif self.flag == Flag.EXCEPTION:
            return (self.body.decode('utf-8'),)

        else:
            raise ValueError(f'{self.flag} is not a valid Flag')

    def is_valid(self, chksum: int):
        '''是否是有效的包体'''
        return self.chksum == chksum


class PacketError(Exception):
    pass


def send_pkt(conn: Connection, packet: Packet):
    '''发送数据报文'''
    datagram = packet.pack()
    conn.sendall(datagram)


def recv_all(conn: Connection, length: int) -> bytes:
    '''接受完整数据'''
    datagram = bytearray()
    while length > 0:
        _data = conn.recv(length)
        n_recv = len(_data)
        if n_recv > 0:
            length -= n_recv
            datagram += _data
        else:
            raise ConnectionResetError

    return bytes(datagram)


def recv_pkt(conn: Connection) -> Packet:
    '''接收数据报文'''
    # 接收并解析 head 部分
    head = recv_all(conn, LEN_HEAD)
    flag, chksum, len_body = Packet.unpack_head(head)

    # 接收 body 部分
    body = recv_all(conn, len_body)
    if crc32(body) == chksum:
        return Packet(flag, body)
    else:
        raise PacketError


class Counter:
    def __init__(self):
        self.n_sent = 0

    def acc(self, length):
        self.n_sent += length


class ConnectionPool(Thread):
    _max_size = 128

    def __init__(self, size=16):
        super().__init__(daemon=True)
        self.size = min(size, self._max_size)
        self.send_q = Queue(self.size)
        self.recv_q = Queue()
        self.done = Event()
        self.sender = SelectSelector()
        self.connections: Set[Connection] = set()

    def send(self, packet: Packet):
        self.send_q.put(packet)

    def recv(self, timeout=TIMEOUT) -> Packet:
        return self.recv_q.get(timeout)

    def add(self, conn: Connection):
        '''添加一个连接'''
        # 检查数量是否达到上限
        if len(self.connections) >= self._max_size:
            return False
        # 检查是否已添加过
        if conn in self.connections:
            return True

        self.connections.add(conn)
        self.sender.register(conn, EVENT_WRITE, data=Counter())

        t_recv = Thread(target=self.listen_to_recv, args=(conn,), daemon=True)
        t_recv.start()
        return True

    def pop(self, conn: Connection):
        try:
            self.sender.unregister(conn)
        except (KeyError, ValueError):
            pass

        try:
            self.connections.remove(conn)
        except KeyError:
            pass
        finally:
            conn.close()

    def listen_to_send(self):
        while not self.done.is_set():
            # find the conn that sent the least data
            keys = [key for key, _ in self.sender.select(timeout=1)]
            if not keys:
                continue
            else:
                key = min(keys, key=lambda k: k.data.n_sent)

            # get data
            packet: Packet = self.send_q.get()
            conn: Connection = key.fileobj

            # send
            try:
                send_pkt(conn, packet)
                key.data.acc(packet.length)
            except SocketError as e:
                self.pop(conn)
                logging.error(f'[Send] conn-{id(conn):x}: {e}, conn exit.')

    def listen_to_recv(self, conn: Connection):
        conn_name = f'{id(conn):x}'
        while not self.done.is_set():
            try:
                packet = recv_pkt(conn)
                self.recv_q.put(packet)
                logging.debug(f'[Recv] conn-{conn_name}: {packet}')
            except ConnectionResetError:
                self.pop(conn)
                return
            except SocketError as e:
                self.pop(conn)
                logging.error(f'[Recv] conn-{conn_name}: {e}, conn exit.')
            except PacketError:
                self.pop(conn)
                logging.error(f'conn-{conn_name} received an error packet.')
                return

    def stop(self):
        self.done.set()
        self.sender.close()
        for conn in self.connections.copy():
            conn.close()

    def run(self):
        if not self.connections:
            raise ValueError('No connection')

        self.done.clear()
        self.listen_to_send()
        self.stop()


class DirInfo:
    '''文件夹信息'''
    __slots__ = ('id', 'perm', 'relpath', 'abspath', '_values')

    def __init__(self, id: int, perm: int, relpath: bytes) -> None:
        self.id = id
        self.perm = perm
        self.relpath = relpath
        self.abspath = Path()

    def __getitem__(self, index):
        if not hasattr(self, '_values'):
            self._values = [self.id, self.perm, self.relpath]
        return self._values[index]

    def __str__(self) -> str:
        return (f'DirInfo(id={self.id}, perm={self.perm}, '
                f'path={self.s_relpath})')

    @classmethod
    def load(cls, dir_id: int, fullpath: Path, relpath: Path):
        d_info = cls(dir_id, fullpath.stat().st_mode, bytes(relpath))
        d_info.abspath = fullpath
        return d_info

    @property
    def s_relpath(self):
        return self.relpath.decode('utf8')

    def set_parent(self, parent: Path):
        '''通过上级目录设置绝对路径'''
        self.abspath = parent.joinpath(self.relpath.decode('utf8'))
        return self.abspath

    def set_stat(self):
        '''设置目录属性'''
        self.abspath.chmod(self.perm)

    def make(self):
        logging.debug(f'[DirInfo] Make dir: {self.s_relpath}')
        self.abspath.mkdir(parents=True, exist_ok=True)
        self.abspath.chmod(self.perm)


class FileInfo:
    '''文件基础信息'''
    __slots__ = ('id', 'perm', 'size', 'mtime', 'chksum', 'relpath', 'abspath',
                 '_values')

    def __init__(self, id: int, perm: int, size: int,
                 mtime: float, chksum: bytes, relpath: bytes):
        self.id = id
        self.perm = perm
        self.size = size
        self.mtime = mtime
        self.chksum = chksum
        self.relpath = relpath
        self.abspath = Path()

    def __getitem__(self, index):
        if not hasattr(self, '_values'):
            self._values = [self.id,
                            self.perm,
                            self.size,
                            self.mtime,
                            self.chksum,
                            self.relpath]
        return self._values[index]

    def __str__(self) -> str:
        return (f'FileInfo(id={self.id}, perm={self.perm:o}, '
                f'sz={self.size}, path={self.s_relpath})')

    @property
    def name(self) -> str:
        return self.abspath.name

    @property
    def n_chunks(self):
        return ceil(self.size / CHUNK_SIZE)

    @classmethod
    def load(cls, file_id: int, fullpath: Path, relpath: Path):
        # 读取文件状态信息
        stat = fullpath.stat()
        f_info = cls(file_id,
                     stat.st_mode,   # 权限, 2 Bytes
                     stat.st_size,   # 大小, 8 Bytes
                     stat.st_mtime,  # 修改时间, 8 Bytes
                     cls.hash(fullpath),  # 文件 MD5 校验码
                     bytes(relpath))
        f_info.abspath = fullpath
        return f_info

    @property
    def s_relpath(self):
        return self.relpath.decode('utf8')

    def set_parent(self, parent: Path):
        '''通过上级目录设置绝对路径'''
        self.abspath = parent.joinpath(self.s_relpath)
        return self.abspath

    def set_stat(self):
        '''设置文件属性'''
        # 设置权限
        self.abspath.chmod(self.perm)
        # 设置时间
        os.utime(self.abspath, (self.mtime, self.mtime))

    def touch(self):
        '''创建空文件'''
        # 确保文件的上级目录存在
        self.abspath.parent.mkdir(mode=0o755, parents=True, exist_ok=True)

        if not self.abspath.exists():
            open(self.abspath, 'w').close()
            self.set_stat()

    def iread(self) -> Generator[Packet, None, None]:
        '''封装文件数据块报文'''
        with open(self.abspath, 'rb') as fp:
            seq = 0
            # 读取单位长度的数据，如果为空则跳出循环
            while True:
                chunk = fp.read(CHUNK_SIZE)
                if chunk:
                    yield Packet.load(Flag.FILE_CHUNK, self.id, seq, chunk)
                    seq += 1
                else:
                    break

    def iwrite(self) -> Generator[None, Tuple[int, bytes], None]:
        '''按数据块迭代写入'''
        # 确保文件的上级目录存在
        self.abspath.parent.mkdir(mode=0o755, parents=True, exist_ok=True)

        # 定义文件所有数据块编号集
        seqs = {i for i in range(self.n_chunks)}

        # 开始迭代写入
        mode = 'rb+' if self.abspath.is_file() else 'wb'
        with open(self.abspath, mode) as fp:
            while seqs:
                seq, chunk = yield
                if seq in seqs:
                    fp.seek(seq * CHUNK_SIZE)
                    fp.write(chunk)
                    seqs.remove(seq)

    @staticmethod
    def hash(filepath: Path) -> bytes:
        hasher = md5()
        with open(filepath, 'rb') as fp:
            while True:
                chunk = fp.read(CHUNK_SIZE)
                if chunk:
                    hasher.update(chunk)
                else:
                    break
        return hasher.digest()

    def is_vaild(self):
        '''检查文件校验和'''
        return (self.abspath.is_file()
                and self.hash(self.abspath) == self.chksum)


class Sender(Thread):
    def __init__(self, sid: bytes, src_paths: List[str], pool_size: int,
                 include=None, exclude=None):
        super().__init__(daemon=True)

        self.sid = sid
        self.srcs = src_paths
        self.conn_pool = ConnectionPool(pool_size)
        self.include = include or '*'
        self.exclude = exclude or []
        self.tree: Dict[int, Union[DirInfo, FileInfo]] = {}

    @staticmethod
    def abspath(path: str):
        if path.startswith('/'):
            return Path(path)
        elif path.startswith('~'):
            return Path(os.path.expanduser(path))
        elif path.startswith('$'):
            return Path(os.path.expandvars(path))
        else:
            return Path.home().joinpath(path)

    @staticmethod
    def traverse_directory(dir_path: Union[str, Path], include):
        '''遍历文件夹'''
        if isinstance(dir_path, str):
            dir_path = Path(dir_path)

        for item in dir_path.rglob(include):
            if item.is_file() or item.is_dir():
                yield item
            else:
                logging.debug(f'[Sender] The `{item}` is not '
                              f'a regular file or dir.')

    @staticmethod
    def need_exclude(path: Path, patterns: Iterable[str]) -> bool:
        for pattern in patterns:
            try:
                if path.match(pattern) or re.search(pattern, path.as_posix()):
                    return True
            except re.error:
                continue
        return False

    @classmethod
    def checkout_paths(cls, fullpath: Path, include: str, exclude: Iterable[str]) \
            -> Generator[Tuple[Path, Path], None, None]:
        '''检出路径'''
        if fullpath.exists():
            if fullpath.is_file():
                relpath = fullpath.relative_to(fullpath.parent)
                if not cls.need_exclude(relpath, exclude):
                    yield fullpath, relpath
            elif fullpath.is_dir():
                for sub_path in cls.traverse_directory(fullpath, include):
                    relpath = sub_path.relative_to(fullpath)
                    if not cls.need_exclude(relpath, exclude):
                        yield sub_path, relpath
            else:
                logging.warning(f'[Sender] The {fullpath} is not '
                                f'a regular file or dir.')
        else:
            logging.warning(f'[Sender] No such file or directory: {fullpath}.')

    @classmethod
    def search_files_and_dirs(cls, path: str, include: str, exclude: list) \
            -> Generator[Tuple[Path, Path], None, None]:
        '''查找文件与文件夹'''
        _path = cls.abspath(path)
        if has_magic(path):
            for matched_path in iglob(str(_path)):
                matched = Path(matched_path)
                for paths in cls.checkout_paths(matched, include, exclude):
                    yield paths
        else:
            for paths in cls.checkout_paths(_path, include, exclude):
                yield paths

    def prepare_all_files(self):
        '''整理要传输的文件列表'''
        _id = 0
        relpaths = set()
        for src_path in self.srcs:
            items = self.search_files_and_dirs(src_path, self.include, self.exclude)
            for fullpath, relpath in items:
                if relpath not in relpaths:
                    # 整理目录树
                    relpaths.add(relpath)
                    if fullpath.is_file():
                        inf_cls, flag = FileInfo, Flag.FILE_INFO
                    else:
                        inf_cls, flag = DirInfo, Flag.DIR_INFO
                    self.tree[_id] = inf_cls.load(_id, fullpath, relpath)

                    # 将 文件/目录 信息发送给接收端
                    info_pkt = Packet.load(flag, *self.tree[_id])
                    self.conn_pool.send(info_pkt)
                    logging.debug(f'[Sender] Found {inf_cls.__name__}: '
                                  f'id={_id} path={relpath.as_posix()}')

                    _id += 1
                else:
                    logging.debug(f'[Sender] Name conflict: '
                                  f'{relpath.as_posix()}, ignore.')

        if _id == 0:
            packet = Packet.load(Flag.EXCEPTION, 'No such file or directory')
        else:
            packet = Packet.load(Flag.FILE_COUNT, _id)
        self.conn_pool.send(packet)
        logging.info(f'[Sender] Num of files and dirs: {_id}')

    def run(self):
        logging.debug(f'[Sender] Sender-{self.sid.hex()[:8]} is running')
        self.conn_pool.start()  # 启动网络连接池

        # 通知对端是否是单文件
        is_monofile = (len(self.srcs) == 1
                       and not has_magic(self.srcs[0])
                       and self.abspath(self.srcs[0]).is_file())
        mono_pkt = Packet.load(Flag.MONOFILE, is_monofile)
        self.conn_pool.send(mono_pkt)

        # 整理所有文件
        Thread(target=self.prepare_all_files, daemon=True).start()

        # 将对端准备就绪的文件读入 output_q
        while True:
            try:
                packet = self.conn_pool.recv()
            except Empty:
                logging.error('[Sender] get input queue timeout, exit.')
                exit_pkt = Packet.load(Flag.EXCEPTION, 'waitting timeout.')
                self.conn_pool.send(exit_pkt)
                break

            if packet.flag == Flag.FILE_READY:
                f_id, = packet.unpack_body()
                f_info = self.tree[f_id]

                # 添加进度条任务
                task_id = progress.add_task(
                    f'upload-{f_info.name}',
                    filename=f_info.name,
                    total=f_info.size,
                    start=True
                )

                # 发送文件数据块
                for chunk_packet in f_info.iread():
                    self.conn_pool.send(chunk_packet)
                    progress.update(task_id, advance=chunk_packet.length)

            elif packet.flag == Flag.DONE:
                logging.info('[Sender] All files are processed, exit.')
                break

            else:
                logging.error(f'[Sender] Unknow packet: {packet}')

        self.conn_pool.stop()
        logging.debug(f'Sender-{self.sid.hex()[:8]} exit')


class Receiver(Thread):
    def __init__(self, sid: bytes, dst_path: str, pool_size: int) -> None:
        super().__init__(daemon=True)

        self.sid = sid
        self.dst_path = Sender.abspath(dst_path)
        self.conn_pool = ConnectionPool(pool_size)

        self.base_dir = Path.home()
        self.size = 0
        self.is_monofile = True
        self.n_recv = 0
        self.total = 0xffffffff
        self.use_custom_name = False
        self.concurrency = Semaphore(8)  # 允许同时写入的文件数
        self.files: Dict[int, FileInfo] = {}
        self.iwriters: Dict[int, Generator] = {}
        self.ready_files: Deque[int] = deque()
        self.progress_tasks: Dict[int, TaskID] = {}

    def check_dst_path(self):
        '''检查目标路径'''
        if self.is_monofile:
            # 单文件传输
            if self.dst_path.is_dir():
                self.base_dir = self.dst_path
            else:
                self.base_dir = self.dst_path.parent
                # 确保保存目录存在
                self.base_dir.mkdir(mode=0o755, parents=True, exist_ok=True)
                self.use_custom_name = True
        else:
            # 多文件传输
            self.base_dir = self.dst_path
            # 确保保存目录存在
            self.base_dir.mkdir(mode=0o755, parents=True, exist_ok=True)

    def process_dir_info(self, packet: Packet):
        '''处理目录信息报文'''
        # 创建目录
        d_info = DirInfo(*packet.unpack_body())
        d_info.set_parent(self.base_dir)
        d_info.make()
        logging.info(f'[Receiver] Dir ready: {d_info}')
        # 接收数量 +1
        self.n_recv += 1

    def ready_notice(self):
        '''通知对端文件准备就绪'''
        while self.ready_files:
            if self.concurrency.acquire(False):
                f_id = self.ready_files[0]
                f_info = self.files[f_id]
                # 创建写入迭代器
                self.iwriters[f_id] = f_info.iwrite()
                self.iwriters[f_id].send(None)

                # 通知对端：文件准备就绪
                logging.debug(f'[Receiver] File({f_id}) ready')
                ready_pkt = Packet.load(Flag.FILE_READY, f_id)
                self.conn_pool.send(ready_pkt)
                self.ready_files.popleft()

                # 添加进度条任务
                task_id = progress.add_task(
                    f'download-{f_info.name}',
                    filename=f_info.name,
                    total=f_info.size,
                    start=True
                )
                self.progress_tasks[f_id] = task_id
            else:
                break

    def process_file_info(self, packet: Packet):
        '''处理文件信息报文'''
        # 解包，并创建 FileInfo 对象
        f_info = FileInfo(*packet.unpack_body())
        if self.use_custom_name:
            f_info.abspath = self.dst_path
        else:
            f_info.set_parent(self.base_dir)

        # 检查文件是否需要传输
        if f_info.is_vaild():
            f_info.set_stat()
            self.n_recv += 1
            logging.info(f'[Receiver] File finished: {f_info.s_relpath}.')
        else:
            if f_info.size > 0:
                self.files[f_info.id] = f_info
                self.size += f_info.size
                self.ready_files.append(f_info.id)  # 将 f_id 加入待通知队列
                self.ready_notice()
            else:
                # 传输的是空文件，直接标记为完成
                f_info.touch()
                self.n_recv += 1
                logging.info(f'[Receiver] File finished: {f_info.s_relpath}')

    def get_iwriter(self, f_id):
        '''获取写入迭代器'''
        if f_id not in self.iwriters:
            f_info = self.files[f_id]
            # 创建并启动写入迭代器
            self.iwriters[f_id] = f_info.iwrite()
            self.iwriters[f_id].send(None)
        return self.iwriters[f_id]

    def process_file_chunk(self, packet: Packet):
        '''处理文件数据块'''
        f_id, seq, chunk = packet.unpack_body()
        try:
            logging.debug(f'[Receiver] Write chunk({seq}) '
                          f'into {self.files[f_id].s_relpath}')
            iwriter = self.get_iwriter(f_id)
            progress.update(self.progress_tasks[f_id], advance=len(chunk))
            iwriter.send((seq, chunk))
        except StopIteration:
            # 释放并发计数器
            self.concurrency.release()
            # 检查文件 Hash
            if self.files[f_id].is_vaild():
                self.files[f_id].set_stat()  # 修改文件状态
                self.n_recv += 1
                self.iwriters.pop(f_id)
                self.ready_notice()
                logging.info(f'[Receiver] File finished: '
                             f'{self.files[f_id].s_relpath}')
            else:
                # TODO: 错误重传
                logging.error(f'[Receiver] Bad file hash: '
                              f'{self.files[f_id].s_relpath}')

        return len(chunk)

    def run(self):
        logging.debug(f'Receiver-{self.sid.hex()[:8]} is running')
        self.conn_pool.start()  # 启动连接池

        # 等待接收文件总数数据包
        logging.debug('[Receiver] Waitting for translation mode')
        packet = self.conn_pool.recv()
        if packet.flag == Flag.MONOFILE:
            # 取出文件总数，并确认目标路径
            self.is_monofile, = packet.unpack_body()
            logging.debug(f'[Receiver] Is monofile: {self.is_monofile}.')
            self.check_dst_path()
        else:
            logging.error(f'[Receiver] The first packet must be `MONOFILE` '
                          f'but receive `{packet.flag.name}`')
            exit_pkt = Packet.load(Flag.EXCEPTION, 'packet type error.')
            self.conn_pool.send(exit_pkt)
            return

        # 等待接收文件信息和数据
        while self.n_recv < self.total:
            packet = self.conn_pool.recv()
            if packet.flag == Flag.DIR_INFO:
                self.process_dir_info(packet)

            elif packet.flag == Flag.FILE_INFO:
                self.process_file_info(packet)

            elif packet.flag == Flag.FILE_CHUNK:
                self.process_file_chunk(packet)

            elif packet.flag == Flag.FILE_COUNT:
                self.total, = packet.unpack_body()

            elif packet.flag == Flag.EXCEPTION:
                msg, = packet.unpack_body()
                logging.error(f'fcp: the sender exit due to `{msg}`')
                break

            else:
                logging.error(f'[Receiver] Unknow packet flag: {packet.flag}')

        self.conn_pool.send(Packet.load(Flag.DONE))
        logging.info('[Receiver] All files finished.')

        self.conn_pool.stop()
        logging.debug(f'Receiver-{self.sid.hex()[:8]} exit')


Porter = Union[Sender, Receiver]



import socket
from selectors import DefaultSelector, EVENT_READ

a, b = socket.socketpair()
b.settimeout(0.001)

c, d = socket.socketpair()
d.settimeout(0.001)

e, f = socket.socketpair()
f.settimeout(0.001)

s = DefaultSelector()
s.register(b, EVENT_READ)
s.register(d, EVENT_READ)
s.register(f, EVENT_READ)


for k, _ in s.select():
    try:
        data = k.fileobj.recv(10, socket.MSG_WAITALL)
        print(k.fd, data)
    except Exception as err:
        print(k.fd, err)
